import enum
import pathlib
from typing import List, Optional

from ilmn.pelops import notifications
from ilmn.pelops.callers import caller_factories
from ilmn.pelops.infrastructure import blacklist_region_repository, pysam_repositories
from ilmn.pelops.interactors import classify_interactor
from ilmn.pelops.ui.cli.presenters import (
    cli_presenter,
    introspection,
    presenter_factories,
)


class InteractorFactory:
    def build_introspection_interactor(
        self,
    ) -> cli_presenter.CliIntrospectionPresenter:
        version_caller = introspection.VersionCaller()
        presenter = cli_presenter.CliIntrospectionPresenter(version_caller)
        return presenter

    def build(
        self,
        bam_file: pathlib.Path,
        output_json: pathlib.Path,
        number_of_threads: int = 1,
        output_dir: Optional[pathlib.Path] = None,
        cli_args: Optional[List[str]] = None,
        bedfile: Optional[pathlib.Path] = None,
        silent: bool = False,
    ) -> classify_interactor.ClassifyInteractor:
        notification_factory = notifications.SimpleNotificationServiceFactory(
            silent=False
        )
        segment_repo_factory = pysam_repositories.PysamSegmentRepositoryFactory(
            bam_file, notification_factory, number_of_threads
        )

        region_repo_factory = blacklist_region_repository.FileRegionRepositoryFactory(
            bedfile
        )
        candidate_region_caller_factory = caller_factories.CandidateRegionCallerFactory(
            region_repo_factory, segment_repo_factory
        )
        region_caller_factory = caller_factories.RegionPairCallerFactory(
            candidate_region_caller_factory, region_repo_factory, notification_factory
        )

        read_caller_factory = caller_factories.ReadCallerFactory(
            segment_repo_factory, notification_factory
        )
        caller_factory = caller_factories.RearrangementCallerFactory(
            segment_repo_factory, region_caller_factory, read_caller_factory
        )
        presenter_factory = presenter_factories.PresenterFactory(
            bam_template_file=bam_file,
            json_output_file=output_json,
            notification_factory=notification_factory,
            cli_args=cli_args,
            output_dir=output_dir,
        )
        presenter = presenter_factory.build_classify_presenter(
            presenter_factories.PresenterType.READ, silent
        )
        result = classify_interactor.ClassifyInteractor(
            presenter, caller_factory, segment_repo_factory
        )
        return result



import abc
import dataclasses
import enum
from typing import (
    Any,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Union,
)


@dataclasses.dataclass
class ClassifiedSegmentCount:
    """Number of observed segments by classification"""

    paired: int
    split: int
    spanning: int


class RegionsName(enum.Enum):
    """Possible names for a CompoundRegions"""

    CoreDUX4 = enum.auto()
    ExtendedDUX4 = enum.auto()
    GRCh38 = enum.auto()
    IGH = enum.auto()
    UNNAMED = enum.auto()
    BLACKLIST = enum.auto()


class GenomicRegion:
    """An immutable genomic region

    1-based genomic region: i.e. ("chr1", 100, 200) means from base 100 to 200
    of contig "chr1" included
    """

    def __init__(self, chrom: str, start: int, end: int):
        self._chrom = chrom
        self._start = start
        self._end = end

    def __eq__(self, other: object) -> bool:
        result = isinstance(other, type(self)) and (
            self.chrom == other.chrom
            and self.start == other.start
            and self.end == other.end
        )
        return result

    def __hash__(self) -> int:
        return hash((self._chrom, self._start, self._end))

    @property
    def chrom(self) -> str:
        return self._chrom

    @property
    def start(self) -> int:
        return self._start

    @property
    def end(self) -> int:
        return self._end

    def overlaps(self, other: object) -> bool:
        """True if self overlaps other. Contiguous regions do not overlap"""
        if not isinstance(other, type(self)) or self.chrom != other.chrom:
            return False

        first, second = sorted([self, other], key=lambda x: x.start)
        return first.end >= second.start

    def within(self, other: object) -> bool:
        """True if self is within other"""
        return (
            isinstance(other, type(self))
            and self.chrom == other.chrom
            and self.start >= other.start
            and self.end <= other.end
        )


class CompoundRegion:
    """A named, immutable set of GenomicRegions."""

    def __init__(self, name: RegionsName, regions: FrozenSet[GenomicRegion]):
        self._name = name
        self._regions = regions

    def __eq__(self, other: object) -> bool:
        result = isinstance(other, type(self)) and (
            self.name == other.name and self.regions == other.regions
        )
        return result

    def __hash__(self) -> int:
        return hash((self._name, self._regions))

    @property
    def name(self) -> RegionsName:
        return self._name

    @property
    def regions(self) -> FrozenSet[GenomicRegion]:
        return self._regions

    def overlaps(self, other: object) -> bool:
        """True if at least one region of self overlaps one region of other.
        Region name is irrelevant"""

        if not isinstance(other, type(self)):
            raise TypeError(
                f"Incompatible type: expected {type(self)}, got {type(other)}"
            )

        for other_region in other.regions:
            if any([region.overlaps(other_region) for region in self.regions]):
                return True

        return False

    def contains(self, other: object) -> bool:
        """True if `other` is within regions of `self`"""
        if isinstance(other, GenomicRegion):
            is_within_genome = any([other.within(region) for region in self._regions])
            return is_within_genome
        raise TypeError(
            f"Incompatible type: expected {type(GenomicRegion)}, got {type(other)}"
        )


class CompoundRegionPair:
    """A pair of CompoundRegions."""

    def __init__(self, a: CompoundRegion, b: CompoundRegion):
        self.__a = a
        self.__b = b

    def __eq__(self, other: object) -> bool:
        result = (
            isinstance(other, type(self)) and self.a == other.a and self.b == other.b
        )
        return result

    def __iter__(self) -> Iterator[CompoundRegion]:
        yield self.__a
        yield self.__b

    @property
    def a(self) -> CompoundRegion:
        return self.__a

    @property
    def b(self) -> CompoundRegion:
        return self.__b

    def get_names(self) -> Tuple[RegionsName, RegionsName]:
        result = self.a.name, self.b.name
        return result


class ReadOrder(enum.Enum):
    """Either read one or read two"""

    ONE = enum.auto()
    TWO = enum.auto()


class PlacedSegment:
    """A Placed Segment that carries over pysam content"""

    def __init__(
        self,
        read_name: str,
        location: FrozenSet[GenomicRegion],
        read_order: ReadOrder,
        content: Any = None,
    ):
        self.__read_name = read_name
        self.__location = location
        self.__read_order = read_order
        self.__content = content

    @property
    def read_name(self) -> str:
        return self.__read_name

    @property
    def location(self) -> FrozenSet[GenomicRegion]:
        return self.__location

    @property
    def read_order(self) -> ReadOrder:
        return self.__read_order

    @property
    def content(self) -> Any:
        """Read content not used by the business logic but required downstream"""
        return self.__content

    def __eq__(self, other: object) -> bool:
        if isinstance(other, type(self)):
            attributes_to_compare = ("read_name", "location", "read_order", "content")
            for attribute in attributes_to_compare:
                if getattr(self, attribute) != getattr(other, attribute):
                    return False
            return True
        else:
            raise TypeError(
                f"Incompatible type: expected {type(self)}, got {type(other)}"
            )

    def __hash__(self) -> int:
        """Uniquely defined by Pysam Segment content"""
        return hash(self.content)


class InvalidReadName(ValueError):
    def __init__(self, read: Union["Read", "ReadPair"], segment: "PlacedSegment"):
        message = f"Impossible to allocate segment {segment.read_name} to read {read.get_name()}"
        super().__init__(message)


class InvalidReadOrder(ValueError):
    def __init__(self, read: Union["Read", "ReadPair"], segment: "PlacedSegment"):
        message = f"Impossible to allocate segment {segment.read_order} to read {read.get_name()}"
        super().__init__(message)


class Read:
    """Models a single read

    A `Read` is created empty and `PlacedSegment` are then allocated
    """

    def __init__(self) -> None:
        self._segments: Set[PlacedSegment] = set()
        self._read_order: Optional["ReadOrder"] = None
        self._read_name: Optional[str] = None

    def __len__(self) -> int:
        return len(self._segments)

    def get_name(self) -> Optional[str]:
        return self._read_name

    def get_locations(self) -> Set[FrozenSet[GenomicRegion]]:
        """Get the locations of the `PlacedSegment` assigned to this read"""
        result = set([segment.location for segment in self._segments])
        return result

    def is_split_across(
        self, a: FrozenSet[GenomicRegion], b: FrozenSet[GenomicRegion]
    ) -> bool:
        if len(self._segments) > 1:
            locations = set([segment.location for segment in self._segments])
            if set([a, b]).issubset(locations):
                return True
        return False

    def allocate(self, placed_segment: PlacedSegment) -> None:
        """Allocate the `PlacedSegment` to the read"""
        if self._read_order is None:
            self._read_order = placed_segment.read_order
        elif self._read_order != placed_segment.read_order:
            raise InvalidReadOrder(self, placed_segment)

        if self._read_name is None:
            self._read_name = placed_segment.read_name
        elif self._read_name != placed_segment.read_name:
            raise InvalidReadName(self, placed_segment)
        self._segments.add(placed_segment)

    def get_segments(self) -> Iterable[PlacedSegment]:
        for segment in self._segments:
            yield segment


class ReadPair:
    """Models a read pair.

    A `ReadPair` is created empty and `PlacedSegment` are then allocated
    """

    def __init__(self) -> None:
        self._pairs: Dict[ReadOrder, Read] = {
            ReadOrder.ONE: Read(),
            ReadOrder.TWO: Read(),
        }
        self._read_name: Optional[str] = None

    def __iter__(self) -> Iterator[Read]:
        yield self._pairs[ReadOrder.ONE]
        yield self._pairs[ReadOrder.TWO]

    def allocate(self, placed_segment: PlacedSegment) -> None:
        """Allocate the placed segment to the correct read of the read pair"""
        if self.get_name() is None:
            self._read_name = placed_segment.read_name
        elif self.get_name() != placed_segment.read_name:
            raise InvalidReadName(self, placed_segment)
        read = self._pairs[placed_segment.read_order]
        read.allocate(placed_segment)

    def get_name(self) -> Optional[str]:
        return self._read_name

    def has_split_read_across(
        self, a: FrozenSet[GenomicRegion], b: FrozenSet[GenomicRegion]
    ) -> bool:
        result = any([read.is_split_across(a, b) for read in self])
        return result

    def has_improper_pair_across(
        self, a: FrozenSet[GenomicRegion], b: FrozenSet[GenomicRegion]
    ) -> bool:
        has_read_across_two_regions = self._has_reads_across(a, b)
        has_split_reads = self.has_split_read_across(a, b)
        result = has_read_across_two_regions and not has_split_reads
        return result

    def has_both_mates(self) -> bool:
        """True if both reads have at least one aligned segment"""
        result = all([len(read) > 0 for read in self._pairs.values()])
        return result

    def get_segments(self) -> Iterable[PlacedSegment]:
        for read in self._pairs.values():
            for segment in read.get_segments():
                yield segment

    def _has_reads_across(
        self, a: FrozenSet[GenomicRegion], b: FrozenSet[GenomicRegion]
    ) -> bool:
        target = set([a, b])
        r1_location, r2_location = [
            target.intersection(read.get_locations()) for read in self
        ]
        return target == r1_location.union(r2_location)


@dataclasses.dataclass
class Rearrangement:
    region_pair: CompoundRegionPair
    counts: ClassifiedSegmentCount
    srpb: float
    segments: Set[PlacedSegment]



import argparse
import pathlib
import warnings
from typing import Any, Dict, List

from ilmn.pelops import defaults, request_models
from ilmn.pelops.factories import interactor_factories


def add_custom_help(parser: argparse.ArgumentParser) -> None:
    """Add a custom help message, so we have control over details"""
    parser.add_argument(
        "--help",
        "-h",
        help="Show this help message and exit.",
        action="help",
        default=argparse.SUPPRESS,
    )


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="pelops",
        description=f"Find spanning reads of genomic rearrangements.",
        add_help=False,
    )
    add_custom_help(parser)
    subparsers = parser.add_subparsers(
        title="Available actions",
        dest="action",
    )

    version_parser = subparsers.add_parser(
        "version",
        help="""Print out package version""",
        add_help=False,
    )
    dux4r_parser = subparsers.add_parser(
        "dux4r",
        help="""Find evidence of DUX4-rearrangements in BAM/CRAM file. Enter `%(prog)s dux4r --help` for more info.""",
        add_help=False,
    )
    populate_classify_parser(dux4r_parser)
    return parser


def populate_classify_parser(classify_parser: argparse.ArgumentParser) -> None:
    classify_parser.add_argument("infile", help="Path to input BAM/CRAM file.")
    add_custom_help(classify_parser)
    classify_parser.add_argument(
        "--json",
        help="Path to the output json file. [DEFAULT=%(default)s]",
        default="pelops_results.json",
        metavar="FILE",
    )
    classify_parser.add_argument(
        "--export",
        help="""Path to the output folder. If provided, supporting reads for each
        rearrangement will be saved in SAM files in such folder.""",
        metavar="DIR",
    )
    classify_parser.add_argument(
        "--total-number-reads",
        type=int,
        help="""Number of reads to use for normalisation.
            If not provided, the total number of unique and mapped reads in the
            BAM file will be used.""",
        metavar="INT",
    )
    classify_parser.add_argument(
        "--threads",
        type=int,
        help="""Number of threads to use when computing total number of reads
        in the input file. [DEFAULT=%(default)s]""",
        metavar="INT",
        default=1,
    )
    classify_parser.add_argument(
        "--srpb-threshold",
        type=float,
        help="""Minimum number of Spanning Read Pairs per Billion required for
        a non-IGH DUX4-rearrangement to be called. [DEFAULT=%(default)s]""",
        default=defaults.srpb_threshold,
        metavar="FLOAT",
    )
    classify_parser.add_argument(
        "--only-igh-dux4",
        help="""If provided, it will only call rearrangements between IGH and
        DUX4 (Core and Extended).""",
        action="store_true",
    )
    classify_parser.add_argument(
        "--minimum-mapq",
        type=int,
        help="""Minimum mapping quality of reads outside of DUX4 region required
            to be counted as spanning read in non-IGH DUX4 rearrangements.
            [DEFAULT=%(default)s]""",
        metavar="INT",
        default=defaults.minimum_mapq,
    )
    classify_parser.add_argument(
        "--filter-regions",
        help="""BED file of regions to ignore when calling non-IGH DUX4-rearrangements.""",
        metavar="FILE",
    )
    classify_parser.add_argument(
        "--silent", help="Disable logging", default=False, action="store_true"
    )
    classify_parser.add_argument(
        "--with-experimental-features",
        help=argparse.SUPPRESS,
        nargs="*",
        default=[],
        choices=[x.value for x in request_models.Feature],
    )


class CliController:
    def __init__(self, interactor_factory: interactor_factories.InteractorFactory):
        self._interactor_factory = interactor_factory

    def dispatch(self, args: List[str]) -> None:
        def custom_warning_formatting(message, category, filename, lineno, file=None, line=None):  # type: ignore
            return f"\n{category.__name__}:{message}\n\n"

        parser = build_parser()
        if len(args) == 1:
            args.append("--help")
        parsed_args = parser.parse_args(args[1:])

        if parsed_args.action == "version":
            introspection = self._interactor_factory.build_introspection_interactor()
            introspection.present_version()
        else:
            interactor_factory_args = self._get_factory_args(parsed_args, args)
            request = self._get_request(parsed_args)
            interactor = self._interactor_factory.build(**interactor_factory_args)
            interactor.present_rearrangement_evidence(request)

    def _get_factory_args(
        self, parsed_args: argparse.Namespace, args: List[str]
    ) -> Dict[str, Any]:
        factory_args: Dict[str, Any] = {}
        factory_args["bam_file"] = pathlib.Path(parsed_args.infile)
        factory_args["output_json"] = pathlib.Path(parsed_args.json)
        factory_args["cli_args"] = args
        factory_args["silent"] = parsed_args.silent
        if getattr(parsed_args, "threads") is not None:
            factory_args["number_of_threads"] = int(parsed_args.threads)
        if getattr(parsed_args, "export"):
            factory_args["output_dir"] = pathlib.Path(parsed_args.export)
        if getattr(parsed_args, "filter_regions") is not None:
            factory_args["bedfile"] = pathlib.Path(parsed_args.filter_regions)
        return factory_args

    def _get_request(
        self, parsed_args: argparse.Namespace
    ) -> request_models.ClassifyRequest:
        features = []
        if not parsed_args.silent:
            features.append(request_models.Feature.WITH_NOTIFICATIONS)
        if not parsed_args.only_igh_dux4:
            features.append(request_models.Feature.DUX4_OTHER)
        if request_models.Feature.DUX4_OTHER in features:
            srpb_threshold = parsed_args.srpb_threshold
            minimum_mapping_quality = parsed_args.minimum_mapq
        else:
            srpb_threshold = minimum_mapping_quality = None

        total_number_of_reads = parsed_args.total_number_reads
        if total_number_of_reads is not None:
            features.append(request_models.Feature.PROVIDED_READ_COUNT)
        if getattr(parsed_args, "filter_regions") is not None:
            features.append(request_models.Feature.WITH_BLACKLIST)
        if parsed_args.with_experimental_features:
            pass  # No experimental features are currently supported
        request = request_models.ClassifyRequest(
            features=frozenset(features),
            minimum_mapping_quality=minimum_mapping_quality,
            srpb_threshold=srpb_threshold,
            total_number_of_reads=total_number_of_reads,
        )
        return request



import abc
from typing import Iterable, Sequence, Set, Tuple

from ilmn.pelops import entities, repositories, selectors
from ilmn.pelops.callers import read_callers, region_pair_callers


class RearrangementCaller(abc.ABC):
    @abc.abstractmethod
    def get_rearrangements(self) -> Iterable[entities.Rearrangement]:
        """Get rearrangement evidence"""


class RegionRearrangementCaller(RearrangementCaller):
    """Finds rearrangements evidence across a region pairs"""

    def __init__(
        self,
        read_caller: read_callers.ReadsCaller,
        region_pair_caller: region_pair_callers.CompoundRegionPairCaller,
        reads_counter: repositories.SegmentCounter,
    ):
        self._read_caller = read_caller
        self._region_pair_caller = region_pair_caller
        self._reads_counter = reads_counter

    def get_rearrangements(self) -> Iterable[entities.Rearrangement]:
        region_pairs = self._region_pair_caller.get_compound_region_pairs()
        for region_pair in region_pairs:
            yield self._get_one_rearrangement(region_pair)

    def _get_one_rearrangement(
        self, region_pair: entities.CompoundRegionPair
    ) -> entities.Rearrangement:
        self._read_caller.detect_reads_spanning_regions(
            region_pair.a.regions, region_pair.b.regions
        )
        segments = set(self._read_caller.get_segments_of_spanning_reads())
        counts = self._read_caller.get_segment_count()
        srpb = (
            counts.spanning
            * 1_000_000_000
            / self._reads_counter.get_number_of_segments()
        )
        result = entities.Rearrangement(region_pair, counts, srpb, segments)
        return result


class MultiRearrangementCaller(RearrangementCaller):
    """A RearrangementCaller made of RearrangementCallers"""

    def __init__(self, callers: Sequence[RearrangementCaller]):
        self._callers = callers

    def get_rearrangements(self) -> Iterable[entities.Rearrangement]:
        for caller in self._callers:
            for rearrangement in caller.get_rearrangements():
                yield rearrangement


class SelectableRearrangementCaller(RearrangementCaller):
    def __init__(
        self, caller: RearrangementCaller, selector: selectors.RearrangementSelector
    ):
        self._caller = caller
        self._selector = selector

    def get_rearrangements(self) -> Iterable[entities.Rearrangement]:
        for item in self._caller.get_rearrangements():
            if self._selector(item):
                yield item


class RearrangementSorter(abc.ABC):
    """Sort Rearrangements"""

    @abc.abstractmethod
    def __call__(
        self, rearrangements: Iterable[entities.Rearrangement]
    ) -> Iterable[entities.Rearrangement]:
        """Sort Rearrangements"""


class SortedRearrangementCaller(RearrangementCaller):
    def __init__(self, caller: RearrangementCaller, sorter: RearrangementSorter):
        self._caller = caller
        self._sorter = sorter

    def get_rearrangements(self) -> Iterable[entities.Rearrangement]:
        for rearrangement in self._sorter(self._caller.get_rearrangements()):
            yield rearrangement



"""Business logic default values"""

# Minimum spanning reads pair per bilion required for non a non-IGH
# DUX4-rearrangement to be called
srpb_threshold = 20.0
# Minimum mapping quality required outside of DUX4 and IGH regions to be
# counted as spanning read
minimum_mapq = 10



import copy
import pathlib
from typing import Any, Dict, Iterable, List, Optional

import pysam

from ilmn.pelops import notifications, result_models
from ilmn.pelops.ui.cli.presenters import cli_presenter


def export_to_sam_file(
    segments: Iterable[pysam.AlignedSegment],
    file_path: pathlib.Path,
    template: Dict[Any, Any],
) -> None:
    outfile = pysam.AlignmentFile(str(file_path), "w", header=template)
    for segment in segments:
        if isinstance(segment, pysam.AlignedSegment):
            outfile.write(segment)
        else:
            raise NotImplementedError(
                f"unable to export segment of type: {type(segment)}"
            )


def create_named_sam_file(
    model_item: cli_presenter.ReadViewModel, output_dir: pathlib.Path
) -> pathlib.Path:
    region_name = "-".join(model_item.region_names)
    file_name = f"{model_item.id}_{region_name}.sam"
    file_path = output_dir / file_name
    return file_path


def find_next_suitable_id(name: str, increment: int, previous_ids: List[str]) -> str:
    while ".".join([name, str(increment)]) in previous_ids:
        increment = increment + 1
    return ".".join([name, str(increment)])


class HeaderManipulator:
    """Manipulate the pysam header to add/edit relevant information"""

    def __init__(self, header: Dict[Any, Any]) -> None:
        self._header = copy.deepcopy(header)  # do not change input header

    def get_header(self) -> Dict[Any, Any]:
        return self._header

    def add_program_details(self, model_item: cli_presenter.ReadViewModel) -> None:
        previous_ids = [program["ID"] for program in self._header["PG"]]
        if model_item.program_name in previous_ids:
            id_ = find_next_suitable_id(model_item.program_name, 1, previous_ids)
        else:
            id_ = model_item.program_name

        program_details = {
            "ID": id_,
            "PN": model_item.program_name,
            "VN": model_item.program_version,
            "CL": model_item.cli_command,
            "DS": f"region_a:{model_item.region_names[0]}  region_b:{model_item.region_names[1]}",
        }
        if len(self._header["PG"]):
            program_details["PP"] = self._header["PG"][-1]["ID"]
        self._header["PG"].append(program_details)


class PysamReadsView(cli_presenter.ReadsView):
    """The Reads View implementation that uses pysam to export reads to file"""

    def __init__(self, template_bam_file: pathlib.Path, output_dir: pathlib.Path):
        self._template = pysam.AlignmentFile(str(template_bam_file), "rb")
        self._output_dir = output_dir

    def update_reads_model(self, model: Iterable[cli_presenter.ReadViewModel]) -> None:
        original_header = self._template.header.to_dict()
        for item in model:
            header_manipulator = HeaderManipulator(original_header)
            header_manipulator.add_program_details(item)
            header = header_manipulator.get_header()
            file_path = create_named_sam_file(item, self._output_dir)
            export_to_sam_file(item.segments, file_path, header)

    def get_ouput_dir(self) -> pathlib.Path:
        return self._output_dir


class NotifyPysamReadsView(cli_presenter.ReadsView):
    def __init__(
        self,
        reads_view: PysamReadsView,
        notification_service: notifications.NotificationService,
    ):
        self.__reads_view = reads_view
        self.__notification_service = notification_service

    def update_reads_model(self, model: Iterable[cli_presenter.ReadViewModel]) -> None:
        self.__reads_view.update_reads_model(model)
        output_dir = self.__reads_view.get_ouput_dir()
        message = f"SAM file exports complete. Results are available in folder '{output_dir}'."
        self.__notification_service.notify(message)



import functools
import logging

from ilmn.pelops import notifications


class LogNotificationService(notifications.NotificationService):
    def notify(self, message: str) -> None:
        logging.info(message)


class LogNotificationServiceFactory(notifications.NotificationServiceFactory):
    @functools.lru_cache(maxsize=1)
    def build(self) -> notifications.NotificationService:
        logging.basicConfig(level=logging.INFO)
        result = LogNotificationService()
        return result



# mypy: no_warn_unused_ignores
import abc
import enum
import functools
import pathlib
from typing import FrozenSet, Iterable, Iterator, List, Optional, Tuple

import pysam

from ilmn.pelops import entities, notifications, repositories
from ilmn.pelops.callers import caller_factories


class SamFlag(enum.IntEnum):
    """All known pysam flags and their numeric representation for sam/bam files"""

    is_paired = 0x1
    is_proper_pair = 0x2
    is_unmapped = 0x4
    mate_is_unmapped = 0x8
    is_reverse = 0x10
    mate_is_reverse = 0x20
    is_read1 = 0x40
    is_read2 = 0x80
    is_secondary = 0x100
    is_qcfail = 0x200
    is_duplicate = 0x400
    is_supplementary = 0x800


class SegmentSelector(abc.ABC):
    @abc.abstractmethod
    def __call__(self, segment: pysam.AlignedSegment) -> bool:
        """Select the segment"""


class PysamPropertyError(TypeError):
    def __init__(self, name: str, value: object, expected: str):
        self.name = name
        self.value = value
        self.expected = expected

    def __str__(self) -> str:
        property_name = f"pysam.AlignedSegment.{self.name}"
        observed = type(self.value).__name__
        return f"{property_name} returned unexpected type {observed}. Expected {self.expected}"


class BooleanSegmentSelector(SegmentSelector):
    """Safely select a pysam.AlignedSegment on property supposed to be boolean"""

    def __init__(self, name: str):
        self.__name = name

    def __call__(self, segment: pysam.AlignedSegment) -> bool:
        """Select on attribute name. Raise a PysamPropertyError if the value is not a boolean"""
        value = getattr(segment, self.__name)
        if not isinstance(value, bool):
            raise PysamPropertyError(self.__name, value, str(bool))
        return value


class SegmentDeSelector(SegmentSelector):
    """Invert the behaviour of a SegmentSelector"""

    def __init__(self, selector: SegmentSelector):
        self.selector = selector

    def __call__(self, segment: pysam.AlignedSegment) -> bool:
        return not self.selector(segment)


class AnySegmentSelector(SegmentSelector):
    """Selector on Any ReadQuery property"""

    def __init__(self, queries: Iterable[repositories.ReadQuery]):
        self._queries = queries
        self.__selector_factory = SelectorFactory()

    def __call__(self, segment: pysam.AlignedSegment) -> bool:
        """If any of the queries selects the segment, the segment is selected"""
        selectors = [self.__selector_factory.build(query) for query in self._queries]
        result = any([selector(segment) for selector in selectors])
        return result


class SelectorFactory:
    """Build SegmentSelector depending on ReadQuery"""

    def build(self, query: repositories.ReadQuery) -> SegmentSelector:
        result: SegmentSelector
        selector_lookup = {
            repositories.ReadQuery.is_duplicate: "is_duplicate",
            repositories.ReadQuery.is_qcfail: "is_qcfail",
            repositories.ReadQuery.is_proper_pair: "is_proper_pair",
            repositories.ReadQuery.is_not_paired: "is_paired",
        }
        deselectors = {repositories.ReadQuery.is_not_paired}
        result = BooleanSegmentSelector(selector_lookup[query])
        if query in deselectors:
            result = SegmentDeSelector(result)
        return result


class MapQSelector(SegmentSelector):
    def __init__(self, minimum_quality: int):
        self._minimum_quality = minimum_quality

    def __call__(self, segment: pysam.AlignedSegment) -> bool:
        """True if segment has at least minimum_quality"""
        mapping_quality = segment.mapping_quality
        if not isinstance(mapping_quality, (int, float)):
            raise PysamPropertyError("mapping_quality", mapping_quality, "number")
        result = segment.mapping_quality >= self._minimum_quality
        return result


class BamFileSegmentCounter(repositories.SegmentCounter):
    default_filters = [
        repositories.ReadQuery.is_unmapped,
        repositories.ReadQuery.is_secondary,
        repositories.ReadQuery.is_qcfail,
        repositories.ReadQuery.is_duplicate,
        repositories.ReadQuery.is_supplementary,
    ]

    lookup = {
        repositories.ReadQuery.is_unmapped: SamFlag.is_unmapped,
        repositories.ReadQuery.is_secondary: SamFlag.is_secondary,
        repositories.ReadQuery.is_qcfail: SamFlag.is_qcfail,
        repositories.ReadQuery.is_duplicate: SamFlag.is_duplicate,
        repositories.ReadQuery.is_supplementary: SamFlag.is_supplementary,
        repositories.ReadQuery.is_proper_pair: SamFlag.is_proper_pair,
    }

    def __init__(self, bam_file: pathlib.Path, number_of_threads: int = 1):
        self._bam_file = bam_file
        self._number_of_threads = number_of_threads

    def get_number_of_segments(
        self, exclude: Optional[List[repositories.ReadQuery]] = None
    ) -> int:
        if exclude is None:
            exclude = self.default_filters
        samflags = [self.lookup[read_query] for read_query in exclude]
        exclude_flag = sum(samflags)

        count: str = pysam.view(  # type: ignore
            "-@",
            f"{self._number_of_threads}",
            "--count",
            "--exclude-flag",
            f"{exclude_flag}",
            str(self._bam_file),
        )
        return int(count.strip())


class PysamMateFinder:
    """Search for mate reads"""

    def __init__(self, bam_file: pysam.AlignmentFile):
        self._bam_file = bam_file

    def get_mate(self, read: pysam.AlignedSegment) -> Optional[pysam.AlignedSegment]:
        if read.is_supplementary:
            result = None
        else:
            try:
                result = self._bam_file.mate(read)
            except ValueError as exc:
                result = self._get_unmapped_mate(read)
        return result

    def _get_unmapped_mate(
        self, read: pysam.AlignedSegment
    ) -> Optional[pysam.AlignedSegment]:
        possible_mates = list(self._get_read_from_mate_position(read))
        if len(possible_mates) == 1:
            return possible_mates[0]
        else:
            # this is unexpected, but we silently carry on
            return None

    def _get_read_from_mate_position(
        self, read: pysam.AlignedSegment
    ) -> Iterator[pysam.AlignedSegment]:
        mate_start = read.next_reference_start
        mate_end = read.next_reference_start + 1
        mate_name = read.next_reference_name
        for candidate in self._bam_file.fetch(
            start=mate_start, stop=mate_end, contig=mate_name, until_eof=True
        ):
            if self._is_mate(read, candidate):
                yield candidate

    def _is_mate(
        self, given: pysam.AlignedSegment, candidate: pysam.AlignedSegment
    ) -> bool:
        result = (
            (candidate.query_name == given.query_name)
            and (given.is_read1 != candidate.is_read1)
            and not candidate.is_supplementary
        )
        return result


class FilePlacedSegmentRepository(repositories.PlacedSegmentRepository):
    _read_map = {True: entities.ReadOrder.ONE, False: entities.ReadOrder.TWO}

    def __init__(
        self,
        bam_file: pathlib.Path,
        read_counter: repositories.SegmentCounter,
    ):
        self._bam_file = pysam.AlignmentFile(str(bam_file), "rb")
        self._mate_finder = PysamMateFinder(self._bam_file)
        self._read_counter = read_counter

    def get_number_of_segments(
        self, exclude: Optional[List[repositories.ReadQuery]] = None
    ) -> int:
        """Count total number of reads, excluding given list of `ReadQuery`"""
        return self._read_counter.get_number_of_segments(exclude)

    def get(
        self,
        locations: FrozenSet[entities.GenomicRegion],
        exclude: List[repositories.ReadQuery],
        min_quality: int = 0,
    ) -> Iterable[entities.PlacedSegment]:
        exclude_selector = AnySegmentSelector(exclude)
        quality_selector = MapQSelector(minimum_quality=min_quality)
        result = (
            self._convert_read(read, locations)
            for read in self._get_reads_from(locations)
            if not exclude_selector(read) and quality_selector(read)
        )
        return result

    def get_mate_exact_position(self, read: entities.PlacedSegment) -> Tuple[str, int]:
        """Get contig name and starting position of read mate"""
        if isinstance(read, entities.PlacedSegment):
            mate_name = str(read.content.next_reference_name)
            mate_start = read.content.next_reference_start + 1
            return mate_name, mate_start
        else:
            raise NotImplementedError(
                f"have not implemented methods to handle read of type: {type(read)}"
            )

    def get_mate(
        self, read: entities.PlacedSegment
    ) -> Optional[entities.PlacedSegment]:
        if not isinstance(read.content, pysam.AlignedSegment):
            raise ValueError("invalid read type")
        mate = self._mate_finder.get_mate(read.content)
        if mate is None:
            result = None
        else:
            result = self._convert_read(mate, frozenset())
        return result

    def _get_reads_from(
        self, locations: FrozenSet[entities.GenomicRegion]
    ) -> Iterator[pysam.AlignedSegment]:
        for region in locations:
            for read in self._bam_file.fetch(region.chrom, region.start, region.end):
                yield read

    def _convert_read(
        self, read: pysam.AlignedSegment, locations: FrozenSet[entities.GenomicRegion]
    ) -> entities.PlacedSegment:
        read_order = self._read_map[read.is_read1]
        if read.query_name is None:
            raise ValueError("AlignedSegment has no name")
        result = entities.PlacedSegment(read.query_name, locations, read_order, read)
        return result


class PysamSegmentRepositoryFactory(repositories.SegmentRepositoryFactory):
    def __init__(
        self,
        bam_file: pathlib.Path,
        notification_factory: notifications.NotificationServiceFactory,
        number_of_threads: int = 1,
    ):
        self.notification_factory = notification_factory
        self._bam_file = bam_file
        self._number_of_threads = number_of_threads

    def build_counter(
        self,
        features: FrozenSet[repositories.SegmentRepoFeature],
        total_number_of_reads: Optional[int],
    ) -> repositories.SegmentCounter:
        if repositories.SegmentRepoFeature.BUILTIN in features:
            if total_number_of_reads is None:
                raise caller_factories.MissingArgumentError("total_number_of_reads")
            else:
                return repositories.ProvidedSegmentCounter(total_number_of_reads)
        else:
            if repositories.SegmentRepoFeature.WITH_NOTIFICATION in features:
                with_notification = True
            else:
                with_notification = False

            return self.__build_singleton_counter(self._bam_file, with_notification)

    @functools.lru_cache(maxsize=1024)
    def __build_singleton_counter(
        self, bam_file: pathlib.Path, with_notification: bool
    ) -> repositories.CachedSegmentCounter:
        """Both the BamFileSegmentCounter and the CachedSegmentCounter need to be unique"""
        result: repositories.SegmentCounter
        result = BamFileSegmentCounter(self._bam_file, self._number_of_threads)
        if with_notification:
            notification_service = self.notification_factory.build()
            result = notifications.NotifySegmentCounter(result, notification_service)
        result = repositories.CachedSegmentCounter(result)
        return result

    def build(
        self,
        features: FrozenSet[repositories.SegmentRepoFeature],
        total_number_of_reads: Optional[int],
    ) -> repositories.PlacedSegmentRepository:
        counter = self.build_counter(features, total_number_of_reads)
        result = FilePlacedSegmentRepository(self._bam_file, counter)
        return result



import abc
import enum
import functools
from typing import FrozenSet, Iterable, List, Optional, Tuple

from ilmn.pelops import entities


class ReadQuery(enum.Enum):
    """Possible queries to filter placed segments"""

    is_duplicate = enum.auto()
    is_not_paired = enum.auto()
    is_proper_pair = enum.auto()
    is_not_proper_pair = enum.auto()
    is_qcfail = enum.auto()
    is_secondary = enum.auto()
    is_supplementary = enum.auto()
    is_unmapped = enum.auto()


class SegmentCounter(abc.ABC):
    @abc.abstractmethod
    def get_number_of_segments(self, exclude: Optional[List[ReadQuery]] = None) -> int:
        """Count total number of reads, excluding given list of `ReadQuery`"""


class ProvidedSegmentCounter(SegmentCounter):
    def __init__(self, provided_number_of_reads: int):
        self._provided_number_of_reads = provided_number_of_reads

    def get_number_of_segments(self, exclude: Optional[List[ReadQuery]] = None) -> int:
        return self._provided_number_of_reads


class CachedSegmentCounter(SegmentCounter):
    def __init__(self, counter: SegmentCounter):
        self.counter = counter

    def get_number_of_segments(self, exclude: Optional[List[ReadQuery]] = None) -> int:
        if exclude is None:
            hashable_exclude = None
        else:
            hashable_exclude = frozenset(exclude)
        return self._get_number_of_segments(hashable_exclude)

    @functools.lru_cache(maxsize=1024)
    def _get_number_of_segments(
        self, hashable_exclude: Optional[FrozenSet[ReadQuery]]
    ) -> int:
        """Results are cached to prevent recomputing very expensive step"""
        if hashable_exclude is None:
            exclude = None
        else:
            exclude = list(hashable_exclude)
        return self.counter.get_number_of_segments(exclude)


class PlacedSegmentRepository(SegmentCounter):
    @abc.abstractmethod
    def get(
        self,
        locations: FrozenSet[entities.GenomicRegion],
        exclude: List[ReadQuery],
        min_quality: int = 0,
    ) -> Iterable[entities.PlacedSegment]:
        """Retrieve segments placed in `location` and cache them"""

    @abc.abstractmethod
    def get_mate(
        self, read: entities.PlacedSegment
    ) -> Optional[entities.PlacedSegment]:
        """Retrieve paired segment of given read"""

    @abc.abstractmethod
    def get_mate_exact_position(self, read: entities.PlacedSegment) -> Tuple[str, int]:
        """Get contig name and starting position of read mate"""


class RegionRepository(abc.ABC):
    @abc.abstractmethod
    def get(self, name: entities.RegionsName) -> entities.CompoundRegion:
        """Get a CompoundRegion by name"""


class BuiltinRegionRepository(RegionRepository):
    _lookup = {
        entities.RegionsName.CoreDUX4: [
            ("chr4", 190020407, 190023665),
            ("chr4", 190066935, 190093279),
            ("chr4", 190172774, 190176845),
            ("chr10", 133663429, 133685936),
            ("chr10", 133739606, 133762125),
        ],
        entities.RegionsName.GRCh38: [
            ("chr1", 1, 248956422),
            ("chr2", 1, 242193529),
            ("chr3", 1, 198295559),
            ("chr4", 1, 190214555),
            ("chr5", 1, 181538259),
            ("chr6", 1, 170805979),
            ("chr7", 1, 159345973),
            ("chr8", 1, 145138636),
            ("chr9", 1, 138394717),
            ("chr10", 1, 133797422),
            ("chr11", 1, 135086622),
            ("chr12", 1, 133275309),
            ("chr13", 1, 114364328),
            ("chr14", 1, 107043718),
            ("chr15", 1, 101991189),
            ("chr16", 1, 90338345),
            ("chr17", 1, 83257441),
            ("chr18", 1, 80373285),
            ("chr19", 1, 58617616),
            ("chr20", 1, 64444167),
            ("chr21", 1, 46709983),
            ("chr22", 1, 50818468),
            ("chrX", 1, 156040895),
            ("chrY", 1, 57227415),
        ],
        entities.RegionsName.ExtendedDUX4: [
            ("chr4", 189967935, 190204560),
            ("chr10", 133564429, 133787422),
            ("chr3", 75667931, 75671185),
            ("chr5", 31248879, 31251987),
            ("chr9", 63816748, 63819462),
            ("chr12", 34207415, 34210675),
            ("chr12", 61599067, 61601843),
            ("chr16", 34134736, 34137792),
            ("chr16", 34140256, 34143849),
            ("chr20", 29317824, 29320579),
            ("chr20", 29323092, 29326049),
            ("chr20", 29409348, 29412600),
            ("chr20", 29447517, 29450306),
            ("chr20", 29877636, 29880363),
            ("chrY", 10170590, 10173725),
            ("chrY", 11305918, 11309181),
            ("chrY", 11313921, 11317187),
            ("chrY", 11320557, 11323823),
            ("chrY", 11331329, 11334595),
        ],
        entities.RegionsName.IGH: [("chr14", 105586937, 106879844)],
    }

    def get(self, name: entities.RegionsName) -> entities.CompoundRegion:
        regions = frozenset(
            entities.GenomicRegion(*region) for region in self._lookup[name]
        )
        result = entities.CompoundRegion(name=name, regions=regions)
        return result


class RegionRepoType(enum.Enum):
    BUILTIN = enum.auto()
    WITH_BLACKLIST = enum.auto()


class RegionRepositoryFactory(abc.ABC):
    @abc.abstractmethod
    def build(self, repo_type: RegionRepoType) -> RegionRepository:
        """Build a RegionRepository"""


class SegmentRepoFeature(enum.Enum):
    WITH_NOTIFICATION = enum.auto()
    BUILTIN = enum.auto()


class SegmentRepositoryFactory(abc.ABC):
    @abc.abstractmethod
    def build_counter(
        self,
        features: FrozenSet[SegmentRepoFeature],
        total_number_of_reads: Optional[int],
    ) -> SegmentCounter:
        """Build a SegmentCounter"""

    @abc.abstractmethod
    def build(
        self,
        features: FrozenSet[SegmentRepoFeature],
        total_number_of_reads: Optional[int],
    ) -> PlacedSegmentRepository:
        """Build a PlacedSegmentRepository"""



import dataclasses
import json
import pathlib
from typing import Dict

from ilmn.pelops import notifications
from ilmn.pelops.ui.cli.presenters import cli_presenter


class JsonClassificationExportView(cli_presenter.ClassificationView):
    def __init__(self, output: pathlib.Path):
        self._output = output

    def update_classification_model(
        self, result: cli_presenter.ClassificationViewModel
    ) -> None:
        with open(self._output, "w") as fh:
            json.dump(dataclasses.asdict(result), fh, indent=4)

    def get_output_file(self) -> pathlib.Path:
        return self._output


class NotifyClassificationView(cli_presenter.ClassificationView):
    def __init__(
        self,
        view_model: JsonClassificationExportView,
        notification_service: notifications.NotificationService,
    ) -> None:
        self.__view_model = view_model
        self.__notification_service = notification_service

    def update_classification_model(
        self, result: cli_presenter.ClassificationViewModel
    ) -> None:
        self.__view_model.update_classification_model(result)
        output_file = self.__view_model.get_output_file()
        message = f"Analysis complete. Results are available in file '{output_file}'."
        self.__notification_service.notify(message)



import abc
from typing import FrozenSet, Iterable, List

from ilmn.pelops import entities, repositories, stores


class ReadsCaller(abc.ABC):
    @abc.abstractmethod
    def detect_reads_spanning_regions(
        self,
        a_regions: FrozenSet[entities.GenomicRegion],
        b_regions: FrozenSet[entities.GenomicRegion],
    ) -> None:
        """Detect reads spanning regions A and B"""

    @abc.abstractmethod
    def get_segment_count(self) -> entities.ClassifiedSegmentCount:
        """Provide detected counts classified segments"""

    @abc.abstractmethod
    def get_segments_of_spanning_reads(self) -> Iterable[entities.PlacedSegment]:
        """Retrieve PlacedSegments from reads spanning across the regions"""


class SpanningReadsCaller(ReadsCaller):
    """Find paired or split reads spanning across two regions"""

    def __init__(
        self,
        placed_segments_repository: repositories.PlacedSegmentRepository,
        reads_to_exclude: List[repositories.ReadQuery],
        minimum_mapping_quality: int = 0,
    ):
        self._placed_segment_repo = placed_segments_repository
        self._reads_to_exclude = reads_to_exclude
        self._minimum_mapping_quality = minimum_mapping_quality
        self._store = stores.PairedReadStore()

    def detect_reads_spanning_regions(
        self,
        a_regions: FrozenSet[entities.GenomicRegion],
        b_regions: FrozenSet[entities.GenomicRegion],
    ) -> None:
        """Detect reads spanning regions A and B, storing them into `PairReadsStore`"""
        self._store.clear()
        for placed_segment in self._placed_segment_repo.get(
            a_regions, exclude=self._reads_to_exclude, min_quality=0
        ):
            try:
                read_pair = self._store.get_read_pair(placed_segment.read_name)
            except stores.UnknowReadPairError:
                read_pair = entities.ReadPair()
            read_pair.allocate(placed_segment)
            self._store.update(read_pair)

        for placed_segment in self._placed_segment_repo.get(
            b_regions,
            exclude=self._reads_to_exclude,
            min_quality=self._minimum_mapping_quality,
        ):
            if placed_segment.read_name in self._store.get_read_names():
                read_pair = self._store.get_read_pair(placed_segment.read_name)
                read_pair.allocate(placed_segment)
                self._store.update(read_pair)

        for read_pair in self._store:
            if read_pair.has_split_read_across(a_regions, b_regions):
                self._store.mark_as_split(read_pair)
                self._store.mark_as_spanning(read_pair)
            if read_pair.has_improper_pair_across(a_regions, b_regions):
                self._store.mark_as_paired(read_pair)
                self._store.mark_as_spanning(read_pair)

    def get_segment_count(self) -> entities.ClassifiedSegmentCount:
        return self._store.get_segment_count()

    def get_segments_of_spanning_reads(self) -> Iterable[entities.PlacedSegment]:
        """Get PlacedSegment that consititute a spanning ReadPair from all spanning ReadPairs"""
        for read_pair in self._store.get_spanning_reads():
            for segment in read_pair.get_segments():
                yield segment
                if not read_pair.has_both_mates():
                    # NOTE: this is a rare event. This happens when the read pair
                    # has a split read, and the mate was not retrieved (e.g. is
                    # unmapped, or mapped outside the pair of CompoundRegion, or
                    # did not pass filters...)
                    mate = self._placed_segment_repo.get_mate(segment)
                    if mate is not None:
                        yield mate



import abc
import enum
from typing import FrozenSet, Optional

from ilmn.pelops import entities, notifications, repositories, selectors
from ilmn.pelops.callers import (
    read_callers,
    rearrangement_callers,
    region_callers,
    region_pair_callers,
    sorters,
)


class MissingArgumentError(ValueError):
    def __init__(self, argument: str):
        message = f"Argument `{argument}` must be provided"
        super().__init__(message)


class CallerType(enum.Enum):
    NAMED = enum.auto()
    BAIT = enum.auto()
    SELECTABLE = enum.auto()
    MULTI = enum.auto()
    SRPB_SORTED = enum.auto()


class CallerFeature(enum.Enum):
    WITH_PROVIDED_READ_COUNT = enum.auto()
    WITH_BLACKLIST = enum.auto()
    WITH_NOTIFICATIONS = enum.auto()


class CandidateRegionCallerFactory:
    def __init__(
        self,
        region_repo_factory: repositories.RegionRepositoryFactory,
        segment_repo_factory: repositories.SegmentRepositoryFactory,
    ):
        self.region_repo_factory = region_repo_factory
        self.segment_repo_factory = segment_repo_factory

    def build(
        self,
        caller_features: FrozenSet[CallerFeature],
        total_number_of_reads: Optional[int] = None,
    ) -> region_callers.CandidateRegionCaller:
        region_repo_type = self.__get_region_repo_type(caller_features)
        selector_features = self.__get_selector_feature(caller_features)
        segment_repo_features = self.__get_segment_repo_features(caller_features)

        region_repo = self.region_repo_factory.build(region_repo_type)

        # TODO: inject this factory, and initialise with another factory, not the region_repository
        selector_factory = selectors.SelectorFactory(region_repo)
        region_selectors = selector_factory.build(features=selector_features)
        segment_repo = self.segment_repo_factory.build(
            segment_repo_features, total_number_of_reads
        )
        result = region_callers.CandidateRegionCaller(segment_repo, region_selectors)
        return result

    def __get_segment_repo_features(
        self, caller_features: FrozenSet[CallerFeature]
    ) -> FrozenSet[repositories.SegmentRepoFeature]:
        result = []
        if CallerFeature.WITH_NOTIFICATIONS in caller_features:
            result.append(repositories.SegmentRepoFeature.WITH_NOTIFICATION)
        if CallerFeature.WITH_PROVIDED_READ_COUNT in caller_features:
            result.append(repositories.SegmentRepoFeature.BUILTIN)
        return frozenset(result)

    def __get_selector_feature(
        self, caller_features: FrozenSet[CallerFeature]
    ) -> FrozenSet[selectors.SelectorFeature]:
        if CallerFeature.WITH_BLACKLIST in caller_features:
            result = frozenset([selectors.SelectorFeature.WITH_BLACKLIST])
        else:
            result = frozenset()
        return result

    def __get_region_repo_type(
        self, caller_features: FrozenSet[CallerFeature]
    ) -> repositories.RegionRepoType:
        if CallerFeature.WITH_BLACKLIST in caller_features:
            result = repositories.RegionRepoType.WITH_BLACKLIST
        else:
            result = repositories.RegionRepoType.BUILTIN
        return result


class ReadCallerFactory:
    __converter = {
        CallerFeature.WITH_NOTIFICATIONS: repositories.SegmentRepoFeature.WITH_NOTIFICATION,
        CallerFeature.WITH_PROVIDED_READ_COUNT: repositories.SegmentRepoFeature.BUILTIN,
    }
    reads_to_exclude = [
        repositories.ReadQuery.is_duplicate,
        repositories.ReadQuery.is_not_paired,
        repositories.ReadQuery.is_qcfail,
    ]

    def __init__(
        self,
        segment_repo_factory: repositories.SegmentRepositoryFactory,
        notification_service_factory: notifications.NotificationServiceFactory,
    ):
        self.segment_repo_factory = segment_repo_factory
        self.notification_service_factory = notification_service_factory

    def build(
        self,
        minimum_mapping_quality: int,
        features: FrozenSet[CallerFeature],
        total_number_of_reads: Optional[int],
    ) -> read_callers.ReadsCaller:
        result: read_callers.ReadsCaller
        repo_features = [
            self.__converter[item] for item in features if item in self.__converter
        ]
        segment_repo = self.segment_repo_factory.build(
            frozenset(repo_features), total_number_of_reads
        )

        result = read_callers.SpanningReadsCaller(
            segment_repo,
            reads_to_exclude=self.reads_to_exclude,
            minimum_mapping_quality=minimum_mapping_quality,
        )
        if CallerFeature.WITH_NOTIFICATIONS in features:
            notification_service = self.notification_service_factory.build()
            result = notifications.NotifyReadsCaller(result, notification_service)
        return result


class RegionPairCallerFactory:
    def __init__(
        self,
        candidate_region_caller_factory: CandidateRegionCallerFactory,
        region_repo_factory: repositories.RegionRepositoryFactory,
        notification_factory: notifications.NotificationServiceFactory,
    ):
        self.candidate_region_caller_factory = candidate_region_caller_factory
        self.region_repo_factory = region_repo_factory
        self.notification_factory = notification_factory

    def build(
        self,
        caller_type: CallerType,
        caller_features: FrozenSet[CallerFeature],
        total_number_of_reads: Optional[int],
    ) -> region_pair_callers.CompoundRegionPairCaller:
        result: region_pair_callers.CompoundRegionPairCaller
        region_repo = self.__get_region_repo(caller_features)

        if caller_type == CallerType.BAIT:
            result = self.__build_bait_reagion_pair_caller(
                region_repo, caller_features, total_number_of_reads
            )
        elif caller_type == CallerType.NAMED:
            result = region_pair_callers.NamedRegionPairCaller(region_repo)
        else:
            raise ValueError()
        if CallerFeature.WITH_NOTIFICATIONS in caller_features:
            notification_service = self.notification_factory.build()
            result = notifications.NotifyRegionPairCaller(result, notification_service)
        return result

    def __get_region_repo(
        self, caller_features: FrozenSet[CallerFeature]
    ) -> repositories.RegionRepository:
        if CallerFeature.WITH_BLACKLIST in caller_features:
            region_repo_type = repositories.RegionRepoType.WITH_BLACKLIST
        else:
            region_repo_type = repositories.RegionRepoType.BUILTIN
        return self.region_repo_factory.build(region_repo_type)

    def __build_bait_reagion_pair_caller(
        self,
        region_repo: repositories.RegionRepository,
        caller_features: FrozenSet[CallerFeature],
        total_number_of_reads: Optional[int],
    ) -> region_pair_callers.BaitRegionPairCaller:
        bait = entities.RegionsName.CoreDUX4
        region_caller = self.candidate_region_caller_factory.build(
            caller_features, total_number_of_reads
        )
        result = region_pair_callers.BaitRegionPairCaller(
            region_repo, region_caller, bait
        )
        return result


class RearrangementCallerFactory:
    def __init__(
        self,
        repo_factory: repositories.SegmentRepositoryFactory,
        region_caller_factory: RegionPairCallerFactory,
        read_caller_factory: ReadCallerFactory,
    ):
        self._repo_factory = repo_factory
        self._region_pair_caller_factory = region_caller_factory
        self._read_caller_factory = read_caller_factory

    def build(
        self,
        caller_type: CallerType,
        features: FrozenSet[CallerFeature],
        minimum_mapping_quality: Optional[int] = None,
        srpb_threshold: Optional[float] = None,
        total_number_of_reads: Optional[int] = None,
    ) -> rearrangement_callers.RearrangementCaller:
        self._srpb_threshold = srpb_threshold
        self._minimum_mapping_quality = minimum_mapping_quality
        return self._build(caller_type, features, total_number_of_reads)

    def _build(
        self,
        caller_type: CallerType,
        caller_features: FrozenSet[CallerFeature],
        total_number_of_reads: Optional[int],
    ) -> rearrangement_callers.RearrangementCaller:
        if caller_type == CallerType.MULTI:
            named_caller = self._build(
                CallerType.NAMED, caller_features, total_number_of_reads
            )
            selectable_caller = self._build(
                CallerType.SRPB_SORTED, caller_features, total_number_of_reads
            )
            callers = [named_caller, selectable_caller]
            return rearrangement_callers.MultiRearrangementCaller(callers)

        elif caller_type == CallerType.SRPB_SORTED:
            sorter = sorters.SrpbRearrangementSorter()
            caller = self._build(
                CallerType.SELECTABLE, caller_features, total_number_of_reads
            )
            return rearrangement_callers.SortedRearrangementCaller(caller, sorter)

        elif caller_type == CallerType.SELECTABLE:
            caller = self._build(
                CallerType.BAIT, caller_features, total_number_of_reads
            )
            if self._srpb_threshold is None:
                raise MissingArgumentError("srpb_threshold")
            selector = selectors.SRPBRearrangementSelector(self._srpb_threshold)
            return rearrangement_callers.SelectableRearrangementCaller(caller, selector)
        else:
            minimum_mapping_quality = self.__get_minimum_mapping_quality(caller_type)
            reads_caller = self._read_caller_factory.build(
                minimum_mapping_quality, caller_features, total_number_of_reads
            )
            region_pair_caller = self._region_pair_caller_factory.build(
                caller_type, caller_features, total_number_of_reads
            )

            segment_repo_features = self.__get_segment_repo_features(caller_features)
            reads_counter = self._repo_factory.build_counter(
                segment_repo_features, total_number_of_reads
            )
            return rearrangement_callers.RegionRearrangementCaller(
                reads_caller, region_pair_caller, reads_counter
            )

    def __get_segment_repo_features(
        self, features: FrozenSet[CallerFeature]
    ) -> FrozenSet[repositories.SegmentRepoFeature]:
        result = []
        if CallerFeature.WITH_PROVIDED_READ_COUNT in features:
            result.append(repositories.SegmentRepoFeature.BUILTIN)
        if CallerFeature.WITH_NOTIFICATIONS in features:
            result.append(repositories.SegmentRepoFeature.WITH_NOTIFICATION)
        return frozenset(result)

    def __get_minimum_mapping_quality(self, caller_type: CallerType) -> int:
        if caller_type == CallerType.NAMED:
            minimum_mapping_quality = 0
        elif caller_type == CallerType.BAIT:
            if self._minimum_mapping_quality is None:
                raise MissingArgumentError("minimum_mapping_quality")
            else:
                minimum_mapping_quality = self._minimum_mapping_quality
        else:
            raise ValueError(f"Unable to build caller for {caller_type.name}")
        return minimum_mapping_quality



from typing import Iterable

from ilmn.pelops import entities
from ilmn.pelops.callers import rearrangement_callers


class SrpbRearrangementSorter(rearrangement_callers.RearrangementSorter):
    def __call__(
        self, rearrangements: Iterable[entities.Rearrangement]
    ) -> Iterable[entities.Rearrangement]:
        """Sort Rearrangements on decreasing SRPB"""
        result = sorted(rearrangements, key=lambda x: x.srpb, reverse=True)
        return iter(result)



import abc
import enum
import pathlib
from typing import FrozenSet, List, Optional

from ilmn.pelops import notifications
from ilmn.pelops.interactors import classify_interactor
from ilmn.pelops.ui.cli.presenters import cli_presenter, introspection
from ilmn.pelops.ui.cli.presenters.views import view_factories


class PresenterType(enum.Enum):
    CLASSIFICATION = enum.auto()
    NULL = enum.auto()
    MULTI = enum.auto()
    READ = enum.auto()


class PresenterFeature(enum.Enum):
    WITH_NOTIFICATIONS = enum.auto()


class PresenterFactory:
    def __init__(
        self,
        bam_template_file: pathlib.Path,
        json_output_file: pathlib.Path,
        notification_factory: notifications.NotificationServiceFactory,
        cli_args: Optional[List[str]],
        output_dir: Optional[pathlib.Path] = None,
    ) -> None:
        self._cli_args = cli_args
        self._output_dir = output_dir
        self.__classify_view_factory = view_factories.ClassifyViewFactory(
            json_output_file, notification_factory
        )
        self.__reads_view_factory = view_factories.ReadsViewFactory(
            bam_template_file, output_dir, notification_factory
        )

    def build_classify_presenter(
        self, presenter_type: PresenterType, silent: bool
    ) -> classify_interactor.ClassifyPresenter:
        classification_presenter = self.build_classification_presenter(silent)
        reads_presenter = self.build_reads_presenter(silent)
        return cli_presenter.MultiPresenter(classification_presenter, reads_presenter)

    def build_classification_presenter(
        self, silent: bool = True
    ) -> cli_presenter.ClassificationPresenter:
        if self._cli_args is None:
            raise ValueError("cli_args must be provided")
        else:
            view_features = self.__get_classify_view_feature(silent)
            classification_view = self.__classify_view_factory.build(view_features)
            cli_introspection = introspection.CliIntrospection(
                introspection.VersionCaller(), self._cli_args
            )
            return cli_presenter.ClassificationPresenter(
                classification_view, cli_introspection
            )

    def build_reads_presenter(
        self, silent: bool = True
    ) -> classify_interactor.ClassifyPresenter:
        if self._output_dir is None:
            return cli_presenter.NullReadPresenter()
        else:
            if self._cli_args is None:
                raise ValueError("cli_args must be provided")
            else:
                version_caller = introspection.VersionCaller()
                cli_introspection = introspection.CliIntrospection(
                    version_caller, cli_args=self._cli_args
                )
                converter = cli_presenter.ReadViewModelConverter(cli_introspection)
                view_features = self.__get_classify_view_feature(silent)
                reads_view = self.__reads_view_factory.build(view_features)
                return cli_presenter.ReadsPresenter(reads_view, converter)

    def __get_classify_view_feature(
        self, silent: bool
    ) -> FrozenSet[view_factories.ClassifyViewFeature]:
        if silent:
            return frozenset()
        else:
            return frozenset([view_factories.ClassifyViewFeature.WITH_NOTIFICATIONS])



import functools
import pathlib
from typing import Iterable, List, Optional

from ilmn.pelops import entities, repositories


class InvalidRegionError(ValueError):
    def __init__(self, region: entities.GenomicRegion) -> None:
        message = f"invalid region: {region.chrom}\t{region.start}-{region.end}"
        super().__init__(message)


class BlackListRegionRepository(repositories.RegionRepository):
    def __init__(
        self, file: pathlib.Path, performs_check_at_init: bool = False
    ) -> None:
        self.__file = file
        self.__builtin_repo = repositories.BuiltinRegionRepository()
        if performs_check_at_init:
            self.get(entities.RegionsName.BLACKLIST)

    @functools.lru_cache(maxsize=None)
    def get(self, name: entities.RegionsName) -> entities.CompoundRegion:
        if name != entities.RegionsName.BLACKLIST:
            return self.__builtin_repo.get(name)
        else:
            regions = self.__extract_regions()
            return entities.CompoundRegion(name=name, regions=frozenset(regions))

    def __extract_regions(self) -> Iterable[entities.GenomicRegion]:
        with open(self.__file, "r") as fh:
            for line in fh:
                if not (line.startswith("#") or line.startswith("track")):
                    columns = self.__split_on_whitespaces(line)
                    region = entities.GenomicRegion(
                        columns[0], int(columns[1]), int(columns[2])
                    )
                    self.__raise_if_invalid_order(region)
                    self.__raise_if_invalid_region(region)
                    yield region

    def __split_on_whitespaces(self, line: str) -> List[str]:
        """split a line on regex [ \t]+"""
        parts = line.strip().replace("\t", " ").split()
        return parts

    def __raise_if_invalid_order(self, region: entities.GenomicRegion) -> None:
        if region.end <= region.start:
            raise InvalidRegionError(region)

    def __raise_if_invalid_region(self, region: entities.GenomicRegion) -> None:
        genome = self.__builtin_repo.get(entities.RegionsName.GRCh38)
        if not genome.contains(region):
            raise InvalidRegionError(region)


class FileRegionRepositoryFactory(repositories.RegionRepositoryFactory):
    def __init__(self, blacklist_bed_file: Optional[pathlib.Path] = None):
        self.__bed_file = blacklist_bed_file

    def build(
        self, repo_type: repositories.RegionRepoType
    ) -> repositories.RegionRepository:
        if repo_type == repositories.RegionRepoType.WITH_BLACKLIST:
            if self.__bed_file is None:
                raise ValueError(
                    "To build blacklist repository a bed file must be provided"
                )
            else:
                return BlackListRegionRepository(
                    self.__bed_file, performs_check_at_init=True
                )
        else:
            return repositories.BuiltinRegionRepository()



import dataclasses
import pathlib
from typing import List

# mypy: no_warn_unused_ignores
try:
    from importlib import metadata as import_metadata  # type: ignore
except ImportError:
    # metadata is added to importlib in CPython 3.8
    # See https://pypi.org/project/importlib-metadata/
    import importlib_metadata as import_metadata  # type: ignore


@dataclasses.dataclass
class EntrypointDetails:
    program_name: str
    program_version: str
    cli_command: str


class VersionCaller:
    """Find the version of this package"""

    def get_version(self) -> str:
        return import_metadata.version("ilmn-pelops")


class CliIntrospection:
    def __init__(self, version_caller: VersionCaller, cli_args: List[str]):
        self._version_caller = version_caller
        self._cli_args = cli_args

    def get_entrypoint_details(self) -> EntrypointDetails:
        version = self._version_caller.get_version()
        command = " ".join(self._cli_args)
        program_name = pathlib.Path(self._cli_args[0]).name
        result = EntrypointDetails(
            program_name=program_name, program_version=version, cli_command=command
        )
        return result



"""Console script for ilmn_pelops."""

import sys
from typing import List

from ilmn.pelops.factories import interactor_factories
from ilmn.pelops.ui.cli import controllers


def main_from_args(args: List[str]) -> int:
    interactor_factory = interactor_factories.InteractorFactory()
    controller = controllers.CliController(interactor_factory)
    controller.dispatch(args)
    return 0


def main() -> int:
    """Console script for ilmn_pelops."""
    return main_from_args(sys.argv)


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



import abc
import enum
from typing import FrozenSet, List, Optional

from ilmn.pelops import entities, repositories


class RearrangementSelector(abc.ABC):
    """Determine if a RearrangementDTO is to be selected or not"""

    @abc.abstractmethod
    def __call__(self, rearrangement: entities.Rearrangement) -> bool:
        """True if this rearrangement is to be selected/kept"""
        pass


class RegionSelector(abc.ABC):
    """Determine if a GenomicRegion is to be selected or not"""

    @abc.abstractmethod
    def __call__(self, region: entities.GenomicRegion) -> bool:
        """True if this region needs to be selected/kept"""


class NamedRegionSelector(RegionSelector):
    """Select a GenomicRegion if it overlaps a named set of GenomicRegions"""

    def __init__(
        self,
        region_name: entities.RegionsName,
        region_repository: repositories.RegionRepository,
    ):
        self._region_name = region_name
        self._repo = region_repository

        # cache regions, so we use the repository only once
        self._regionset: Optional[entities.CompoundRegion] = None

    def __call__(self, region: entities.GenomicRegion) -> bool:
        if not self._regionset:
            self._regionset = self._repo.get(self._region_name)
        result = any([region.overlaps(target) for target in self._regionset.regions])
        return result


class RegionDeSelector(RegionSelector):
    """Inverts the behaviour of a RegionSelector"""

    def __init__(self, selector: RegionSelector):
        self._selector = selector

    def __call__(self, region: entities.GenomicRegion) -> bool:
        result = not self._selector(region)
        return result


class SRPBRearrangementSelector(RearrangementSelector):
    """Determine if a RearrangementDTO is to be selected or not based on SRPB"""

    def __init__(self, threshold: float):
        self._threshold = threshold

    def __call__(self, rearrangement: entities.Rearrangement) -> bool:
        return rearrangement.srpb >= self._threshold


class SelectorFeature(enum.Enum):
    WITH_BLACKLIST = enum.auto()


class SelectorFactory:
    def __init__(self, regions_repository: repositories.RegionRepository):
        self._regions_repository = regions_repository

    def build(
        self, features: FrozenSet[SelectorFeature] = frozenset()
    ) -> List[RegionSelector]:
        result = [
            NamedRegionSelector(entities.RegionsName.GRCh38, self._regions_repository),
            RegionDeSelector(
                NamedRegionSelector(
                    entities.RegionsName.ExtendedDUX4, self._regions_repository
                )
            ),
            RegionDeSelector(
                NamedRegionSelector(entities.RegionsName.IGH, self._regions_repository)
            ),
        ]
        if SelectorFeature.WITH_BLACKLIST in features:
            result.append(
                RegionDeSelector(
                    NamedRegionSelector(
                        entities.RegionsName.BLACKLIST, self._regions_repository
                    )
                )
            )

        return result



"""Module dealing with compound regions handled in pairs."""

import abc
from typing import Iterable

from ilmn.pelops import entities
from ilmn.pelops.callers import region_callers
from ilmn.pelops.repositories import RegionRepository


class CompoundRegionPairCaller(abc.ABC):
    """Interface for CompoundRegionPairCaller.

    Classes in this hierarchy identify pairs of compound regions according to
    various criteria.

    """

    @abc.abstractmethod
    def get_compound_region_pairs(self) -> Iterable[entities.CompoundRegionPair]:
        """Get Pairs of CompoundRegion"""


class NamedRegionPairCaller(CompoundRegionPairCaller):
    """Find CompoundRegion pairs"""

    name_pairs = [
        (entities.RegionsName.CoreDUX4, entities.RegionsName.IGH),
        (entities.RegionsName.ExtendedDUX4, entities.RegionsName.IGH),
    ]

    def __init__(self, region_repository: RegionRepository):
        self._region_repo = region_repository

    def get_compound_region_pairs(self) -> Iterable[entities.CompoundRegionPair]:
        """Convert tuples of `RegionName` to tuples of `CompoundRegion`"""
        for pair in self.name_pairs:
            result = entities.CompoundRegionPair(
                self._region_repo.get(pair[0]),
                self._region_repo.get(pair[1]),
            )
            yield result


class BaitRegionPairCaller(CompoundRegionPairCaller):
    """Find CompoundRegion pairs by looking for candidates to a specific
    "bait" region."""

    def __init__(
        self,
        region_repository: RegionRepository,
        region_caller: region_callers.CandidateRegionCaller,
        bait: entities.RegionsName,
    ):
        self._region_repo = region_repository
        self._region_caller = region_caller
        self._bait = bait

    def get_compound_region_pairs(self) -> Iterable[entities.CompoundRegionPair]:
        bait_compound_region = self._region_repo.get(self._bait)
        for candidate_region in self._region_caller.get_candidates_regions(
            bait_compound_region
        ):
            candidate_compound_region = entities.CompoundRegion(
                entities.RegionsName.UNNAMED, frozenset([candidate_region])
            )

            result = entities.CompoundRegionPair(
                bait_compound_region, candidate_compound_region
            )

            yield result



import itertools
from typing import Iterable, Iterator, List, Tuple

from ilmn.pelops import entities, repositories, selectors, stores

GenomicRegion = entities.GenomicRegion


def pairwise(
    iterable: Iterable[GenomicRegion],
) -> Iterator[Tuple[GenomicRegion, GenomicRegion]]:
    """Return successive overlapping pairs taken from iterable."""
    # part of itertools only from 3.10. Using their documented equivalent behaviour
    # https://docs.python.org/3/library/itertools.html#itertools.pairwise
    # pairwise('ABCDEFG') --> AB BC CD DE EF FG
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


class ContiguousRegionsCaller:
    """Find contiguous GenomicRegions"""

    def __init__(self) -> None:
        self._current_block_of_contiguous_regions: List[GenomicRegion] = []

    def find_contiguous(
        self, regions: Iterable[GenomicRegion]
    ) -> Iterable[List[GenomicRegion]]:
        """Go through an iterable of sorted `GenomicRegion` and returns
        an iterable of blocks of two or more contiguous `GenomicRegion`"""

        for previous, current in pairwise(regions):
            if self._are_contiguous(previous, current):
                self._add_to_block(previous, current)
            elif self._current_block_of_contiguous_regions:
                yield self._current_block_of_contiguous_regions
                self._current_block_of_contiguous_regions = []

        # at the end we want to return contiguous regions
        if len(self._current_block_of_contiguous_regions) > 1:
            yield self._current_block_of_contiguous_regions

    def _add_to_block(self, previous: GenomicRegion, current: GenomicRegion) -> None:
        if not self._current_block_of_contiguous_regions:
            self._current_block_of_contiguous_regions.append(previous)
        self._current_block_of_contiguous_regions.append(current)

    def _are_contiguous(self, a: GenomicRegion, b: GenomicRegion) -> bool:
        return a.chrom == b.chrom and a.end + 1 >= b.start


class RegionConsolidationService:
    """Merge or remove regions"""

    low_pass_filter = 2

    def __init__(self) -> None:
        self._store = stores.RegionStore()
        self._contiguous_region_caller = ContiguousRegionsCaller()

    def add(self, candidate: GenomicRegion) -> None:
        self._store.add(candidate)

    def get_regions(self) -> Iterable[GenomicRegion]:
        for region, count in self._store.get_counted_regions():
            yield region

    def consolidate(self) -> None:
        self._remove_low_count_regions()
        self._merge_consecutive_regions()

    def _merge_consecutive_regions(self) -> None:
        """Replace contiguous regions with a single merged region"""

        regions = self._store.get_regions()
        region_blocks = self._contiguous_region_caller.find_contiguous(regions)
        for contiguous_regions in region_blocks:
            merged_region = self._merge(contiguous_regions)
            self._store.add(merged_region)
            for region in contiguous_regions:
                self._store.delete(region)

    def _merge(self, group: List[GenomicRegion]) -> GenomicRegion:
        previous = group[0]
        for current in group[1:]:
            previous = entities.GenomicRegion(
                previous.chrom, previous.start, current.end
            )
        return previous

    def _remove_low_count_regions(self) -> None:
        for region, count in self._store.get_counted_regions():
            if count < self.low_pass_filter:
                self._store.delete(region)


class CandidateRegionCaller:
    """Find candidate mate GenomicRegion(s) of a given CompoundRegion.

    Given a CompoundRegion, finds candidate GenomicRegion(s) where a
    rearrangement might be occurring, based on evidence from placed segments"""

    region_size = 1000
    reads_to_exclude = [
        repositories.ReadQuery.is_duplicate,
        repositories.ReadQuery.is_not_paired,
        repositories.ReadQuery.is_proper_pair,
        repositories.ReadQuery.is_qcfail,
    ]

    def __init__(
        self,
        segment_repo: repositories.PlacedSegmentRepository,
        selectors: List[selectors.RegionSelector],
    ):
        self._segment_repo = segment_repo
        self._candidates = RegionConsolidationService()
        self._selectors = selectors

    def get_candidates_regions(
        self, origin: entities.CompoundRegion
    ) -> Iterable[GenomicRegion]:
        for segment in self._segment_repo.get(
            origin.regions, exclude=self.reads_to_exclude
        ):
            chrom, pos = self._segment_repo.get_mate_exact_position(segment)
            start, end = self._get_boundaries(pos)
            region = GenomicRegion(chrom, start, end)
            if all([selector(region) for selector in self._selectors]):
                self._candidates.add(region)
        self._candidates.consolidate()
        candidates = set(self._candidates.get_regions())
        return iter(candidates)

    def _get_boundaries(self, pos: int) -> Tuple[int, int]:
        low = int(pos / self.region_size) * self.region_size
        high = low + self.region_size
        return low + 1, high



"""Result Model for classify"""

import dataclasses
import enum
from typing import Any, FrozenSet, List, Optional, Set, Tuple

# NOTE: order of class definition in the file matters. Early class definition
# should not contain classes defined below itself, so we can refer to them by
# class name rather than by string


class ReferenceGenome(enum.Enum):
    GRCh38 = enum.auto()


@dataclasses.dataclass(frozen=True)
class GenomicRegionDTO:
    chrom: str
    start: int
    end: int


@dataclasses.dataclass(frozen=True)
class CompoundRegionDTO:
    name: str
    regions: FrozenSet[GenomicRegionDTO]


@dataclasses.dataclass(frozen=True)
class ReadsEvidence:
    paired: int
    split: int
    SRPB: float  # spanning read pairs per billion


@dataclasses.dataclass(frozen=True)
class PlacedSegmentDTO:
    read_name: str
    content: Any = None


@dataclasses.dataclass(frozen=True)
class RearrangementDTO:
    A: CompoundRegionDTO
    B: CompoundRegionDTO
    evidence: ReadsEvidence
    supporting_reads: Set[PlacedSegmentDTO]


@dataclasses.dataclass
class ClassifyResult:
    reference: ReferenceGenome
    unique_mapped_reads: int
    rearrangements: List[RearrangementDTO]



"""Interactors covering all usecases for pelops"""

import abc
from typing import FrozenSet, List

from ilmn.pelops import entities, repositories, request_models, result_models
from ilmn.pelops.callers import caller_factories


class ClassifyPresenter(abc.ABC):
    @abc.abstractmethod
    def present_classification(self, result: result_models.ClassifyResult) -> None:
        """Present the result from classify_reads"""


def convert_compound_region(
    domain: entities.CompoundRegion,
) -> result_models.CompoundRegionDTO:
    """Convert the domain object `CompoundRegion` into a data structure"""
    regions = [convert_genomic_region(item) for item in domain.regions]
    result = result_models.CompoundRegionDTO(
        name=domain.name.name, regions=frozenset(regions)
    )
    return result


def convert_genomic_region(
    domain: entities.GenomicRegion,
) -> result_models.GenomicRegionDTO:
    """Convert the domain object `GenomicRegionDTO` into a data structure"""
    result = result_models.GenomicRegionDTO(
        chrom=domain.chrom, start=domain.start, end=domain.end
    )
    return result


def convert_segment(domain: entities.PlacedSegment) -> result_models.PlacedSegmentDTO:
    result = result_models.PlacedSegmentDTO(
        read_name=domain.read_name, content=domain.content
    )
    return result


def convert_rearrangement(
    rearrangement: entities.Rearrangement,
) -> result_models.RearrangementDTO:
    evidence = result_models.ReadsEvidence(
        rearrangement.counts.paired, rearrangement.counts.split, rearrangement.srpb
    )
    segments = [convert_segment(item) for item in rearrangement.segments]
    a, b = rearrangement.region_pair
    result = result_models.RearrangementDTO(
        convert_compound_region(a),
        convert_compound_region(b),
        evidence,
        set(segments),
    )
    return result


class ClassifyInteractor:
    def __init__(
        self,
        presenter: ClassifyPresenter,
        rearrangement_caller_factory: caller_factories.RearrangementCallerFactory,
        repository_factory: repositories.SegmentRepositoryFactory,
    ):
        self._presenter = presenter
        self._caller_factory = rearrangement_caller_factory
        self._repo_factory = repository_factory

    def get_rearrangement_evidence(
        self, request: request_models.ClassifyRequest
    ) -> result_models.ClassifyResult:
        """Classify reads spanning several GenomiRegionSets pairs into
        spanning, paired, and split reads."""
        rearrangement_caller = self._caller_factory.build(
            self.__get_caller_type(request.features),
            self.__get_caller_features(request.features),
            request.minimum_mapping_quality,
            request.srpb_threshold,
            request.total_number_of_reads,
        )

        reads_counter = self._repo_factory.build_counter(
            self.__get_counter_features(request.features),
            request.total_number_of_reads,
        )

        rearrangements = [
            convert_rearrangement(item)
            for item in rearrangement_caller.get_rearrangements()
        ]
        result = result_models.ClassifyResult(
            reference=result_models.ReferenceGenome.GRCh38,
            unique_mapped_reads=reads_counter.get_number_of_segments(),
            rearrangements=list(rearrangements),
        )
        return result

    def present_rearrangement_evidence(
        self, request: request_models.ClassifyRequest
    ) -> None:
        result = self.get_rearrangement_evidence(request)
        self._presenter.present_classification(result)

    def __get_caller_type(
        self, features: FrozenSet[request_models.Feature]
    ) -> caller_factories.CallerType:
        if request_models.Feature.DUX4_OTHER in features:
            return caller_factories.CallerType.MULTI
        else:
            return caller_factories.CallerType.NAMED

    def __get_caller_features(
        self, features: FrozenSet[request_models.Feature]
    ) -> FrozenSet[caller_factories.CallerFeature]:
        converter = {
            request_models.Feature.PROVIDED_READ_COUNT: caller_factories.CallerFeature.WITH_PROVIDED_READ_COUNT,
            request_models.Feature.WITH_BLACKLIST: caller_factories.CallerFeature.WITH_BLACKLIST,
            request_models.Feature.WITH_NOTIFICATIONS: caller_factories.CallerFeature.WITH_NOTIFICATIONS,
        }
        result = [converter[feature] for feature in features if feature in converter]
        return frozenset(result)

    def __get_counter_features(
        self, features: FrozenSet[request_models.Feature]
    ) -> FrozenSet[repositories.SegmentRepoFeature]:
        converter = {
            request_models.Feature.WITH_NOTIFICATIONS: repositories.SegmentRepoFeature.WITH_NOTIFICATION,
            request_models.Feature.PROVIDED_READ_COUNT: repositories.SegmentRepoFeature.BUILTIN,
        }
        result = [converter[feature] for feature in features if feature in converter]
        return frozenset(result)



import dataclasses
import enum
from typing import FrozenSet, Optional


class Feature(enum.Enum):
    DUX4_OTHER = enum.auto()
    PROVIDED_READ_COUNT = enum.auto()
    WITH_BLACKLIST = enum.auto()
    WITH_NOTIFICATIONS = enum.auto()


@dataclasses.dataclass
class ClassifyRequest:
    features: FrozenSet[Feature] = frozenset()
    minimum_mapping_quality: Optional[int] = None
    srpb_threshold: Optional[float] = None
    total_number_of_reads: Optional[int] = None



import abc
import functools
from typing import FrozenSet, Iterable, List, Optional

from ilmn.pelops import entities, repositories
from ilmn.pelops.callers import read_callers, region_pair_callers


class NotificationService(abc.ABC):
    """A service to dipatch notification"""

    @abc.abstractmethod
    def notify(self, message: str) -> None:
        """Dispatch one notification"""


class NotificationServiceFactory(abc.ABC):
    @abc.abstractmethod
    def build(self) -> NotificationService:
        """Build notification service"""


class SimpleNotificationService(NotificationService):
    """A simple notification service that uses native `print` to notify"""

    def __init__(self, silent: bool = True) -> None:
        self.__store: List[str] = []
        self.__silent = silent

    def notify(self, message: str) -> None:
        if not self.__silent:
            print(message)
        self.__store.append(message)

    def get_notifications(self) -> List[str]:
        """Retrieve all the notification received so far"""
        return self.__store


class SimpleNotificationServiceFactory(NotificationServiceFactory):
    def __init__(self, silent: bool = True):
        self.__silent = silent

    @functools.lru_cache(maxsize=1)
    def build(self) -> NotificationService:
        return SimpleNotificationService(self.__silent)


class NotifySegmentCounter(repositories.SegmentCounter):
    """SegmentCounter composed with NotificationService"""

    def __init__(
        self,
        read_counter: repositories.SegmentCounter,
        notification_service: NotificationService,
    ):
        self.__read_counter = read_counter
        self.__notification_service = notification_service

    def get_number_of_segments(
        self, exclude: Optional[List[repositories.ReadQuery]] = None
    ) -> int:
        message = "Counting number of unique and mapped reads."
        self.__notification_service.notify(message)
        return self.__read_counter.get_number_of_segments(exclude)


class NotifyReadsCaller(read_callers.ReadsCaller):
    def __init__(
        self,
        read_caller: read_callers.ReadsCaller,
        notification_service: NotificationService,
    ):
        self.read_caller = read_caller
        self.notification_service = notification_service

    def detect_reads_spanning_regions(
        self,
        a_regions: FrozenSet[entities.GenomicRegion],
        b_regions: FrozenSet[entities.GenomicRegion],
    ) -> None:
        self.notification_service.notify("Finding evidence for rearrangement.")
        self.read_caller.detect_reads_spanning_regions(a_regions, b_regions)

    def get_segment_count(self) -> entities.ClassifiedSegmentCount:
        return self.read_caller.get_segment_count()

    def get_segments_of_spanning_reads(self) -> Iterable[entities.PlacedSegment]:
        return self.read_caller.get_segments_of_spanning_reads()


class NotifyRegionPairCaller(region_pair_callers.CompoundRegionPairCaller):
    def __init__(
        self,
        region_pair_caller: region_pair_callers.CompoundRegionPairCaller,
        notification_service: NotificationService,
    ):
        self.__region_pair_caller = region_pair_caller
        self.__notification_service = notification_service

    def get_compound_region_pairs(self) -> Iterable[entities.CompoundRegionPair]:
        """Report about compund region pairs retrived"""
        for region_pair in self.__region_pair_caller.get_compound_region_pairs():
            A, B = (item.name for item in region_pair.get_names())
            if B == entities.RegionsName.UNNAMED.name:
                B = "candidate-region"
            message = f"Retrieving region coordinates for {A}/{B}."
            self.__notification_service.notify(message)
            yield region_pair



import abc
import dataclasses
from typing import Iterable, Tuple

import pysam

from ilmn.pelops import result_models
from ilmn.pelops.interactors import classify_interactor
from ilmn.pelops.ui.cli.presenters import introspection


@dataclasses.dataclass
class ReadViewModel:
    id: str
    region_names: Tuple[str, str]
    segments: Iterable[pysam.AlignedSegment]
    program_name: str
    program_version: str
    cli_command: str


class ReadsView(abc.ABC):
    @abc.abstractmethod
    def update_reads_model(self, model: Iterable[ReadViewModel]) -> None:
        """Update the reads view model"""


class Formatter:
    """Format number coherently"""

    def format_id(self, id_: int) -> str:
        result = format(id_, "02d")
        return result


class ReadViewModelConverter:
    def __init__(
        self,
        cli_introspection: introspection.CliIntrospection,
    ):
        self._formatter = Formatter()
        self._cli_introspection = cli_introspection

    def convert(
        self, i: int, rearrangement: result_models.RearrangementDTO
    ) -> ReadViewModel:
        id_ = self._formatter.format_id(i)
        region_names = (rearrangement.A.name, rearrangement.B.name)
        sorted_segments = sorted(
            rearrangement.supporting_reads, key=lambda segment: segment.read_name
        )
        sorted_reads = [item.content for item in sorted_segments]

        entrypoint_details = self._cli_introspection.get_entrypoint_details()
        result = ReadViewModel(
            id=id_,
            region_names=region_names,
            segments=sorted_reads,
            program_name=entrypoint_details.program_name,
            program_version=entrypoint_details.program_version,
            cli_command=entrypoint_details.cli_command,
        )
        return result


class ReadsPresenter(classify_interactor.ClassifyPresenter):
    def __init__(
        self,
        reads_view: ReadsView,
        converter: ReadViewModelConverter,
    ):
        self._reads_view = reads_view
        self._converter = converter

    def present_classification(self, result: result_models.ClassifyResult) -> None:
        view_model = [
            self._converter.convert(i + 1, rearrangement)
            for i, rearrangement in enumerate(result.rearrangements)
        ]
        self._reads_view.update_reads_model(view_model)


class NullReadPresenter(classify_interactor.ClassifyPresenter):
    """The Reads Presenter implementation that does not present reads"""

    def __init__(self) -> None:
        pass

    def present_classification(self, result: result_models.ClassifyResult) -> None:
        pass


@dataclasses.dataclass
class ViewEvidence:
    paired_reads: int
    split_reads: int
    SRPB: float  # spanning read pairs per billion


@dataclasses.dataclass
class ViewRegion:
    chrom: str
    start: int
    end: int


@dataclasses.dataclass
class ViewCompoundRegion:
    name: str
    regions: Iterable[ViewRegion]


@dataclasses.dataclass
class ViewRearrangement:
    id: str
    A: ViewCompoundRegion
    B: ViewCompoundRegion
    evidence: ViewEvidence


@dataclasses.dataclass
class ClassificationViewModel:
    reference: str
    unique_mapped_reads: int
    rearrangements: Iterable[ViewRearrangement]
    program_name: str
    version: str
    cli_command: str


class ClassificationView(abc.ABC):
    @abc.abstractmethod
    def update_classification_model(self, result: ClassificationViewModel) -> None:
        """Update the classification view model"""


class ResultModelConverter:
    """Convert part of the result model to part of the view model"""

    def __init__(self) -> None:
        self._formatter = Formatter()

    def convert_result_model(
        self,
        result: result_models.ClassifyResult,
        cli_details: introspection.EntrypointDetails,
    ) -> ClassificationViewModel:
        reference = result.reference.name
        unique_mapped_reads = result.unique_mapped_reads
        rearrangements = [
            self.convert_rearrangement(i + 1, item)
            for i, item in enumerate(result.rearrangements)
        ]
        view_model = ClassificationViewModel(
            reference,
            unique_mapped_reads,
            rearrangements,
            program_name=cli_details.program_name,
            version=cli_details.program_version,
            cli_command=cli_details.cli_command,
        )
        return view_model

    def convert_rearrangement(
        self, i: int, rearrangement: result_models.RearrangementDTO
    ) -> "ViewRearrangement":
        id_ = self._formatter.format_id(i)
        A = self.convert_genomic_region_set(rearrangement.A)
        B = self.convert_genomic_region_set(rearrangement.B)
        evidence = self.convert_evidence(rearrangement.evidence)
        result = ViewRearrangement(id_, A, B, evidence)
        return result

    def convert_genomic_region_set(
        self, region_set: result_models.CompoundRegionDTO
    ) -> "ViewCompoundRegion":
        regions = [
            ViewRegion(**dataclasses.asdict(region)) for region in region_set.regions
        ]
        result = ViewCompoundRegion(region_set.name, regions)
        return result

    def convert_evidence(self, evidence: result_models.ReadsEvidence) -> "ViewEvidence":
        SRPB = round(evidence.SRPB, 2)
        result = ViewEvidence(evidence.paired, evidence.split, SRPB)
        return result


class ClassificationPresenter(classify_interactor.ClassifyPresenter):
    def __init__(
        self,
        classification_view: ClassificationView,
        cli_introspection: introspection.CliIntrospection,
    ) -> None:
        self._classification_view = classification_view
        self._cli_introspection = cli_introspection
        self._converter = ResultModelConverter()

    def present_classification(self, result: result_models.ClassifyResult) -> None:
        entrypoint_details = self._cli_introspection.get_entrypoint_details()
        view_model = self._converter.convert_result_model(result, entrypoint_details)
        self._classification_view.update_classification_model(view_model)


class MultiPresenter(classify_interactor.ClassifyPresenter):
    """A concrete Presenter which delegates"""

    def __init__(
        self,
        classification_presenter: classify_interactor.ClassifyPresenter,
        reads_presenter: classify_interactor.ClassifyPresenter,
    ):
        self._classification_presenter = classification_presenter
        self._reads_presenter = reads_presenter

    def present_classification(self, result: result_models.ClassifyResult) -> None:
        """Present the result from classify_reads"""
        self._classification_presenter.present_classification(result)
        self._reads_presenter.present_classification(result)


class CliIntrospectionPresenter:
    def __init__(self, version_caller: introspection.VersionCaller) -> None:
        self.__version_caller = version_caller

    def present_version(self) -> None:
        print(self.__version_caller.get_version())



import enum
import pathlib
from typing import FrozenSet, Optional

from ilmn.pelops import notifications
from ilmn.pelops.ui.cli.presenters import cli_presenter
from ilmn.pelops.ui.cli.presenters.views import (
    json_classification_export_view,
    pysam_segment_export_view,
)


class ResultFileExistsError(FileExistsError):
    pass


class InvalidDirectoryError(IOError):
    pass


def _raise_missing_directory(directory: pathlib.Path) -> None:
    message = f"Directory {directory} does not exist. Aborting"
    raise InvalidDirectoryError(message)


def _check_output(output_json: pathlib.Path) -> None:
    parent_directory = output_json.parent
    if output_json.exists():
        message = f"Result file {output_json} already exists. Aborting"
        raise ResultFileExistsError(message)
    elif not parent_directory.exists():
        _raise_missing_directory(parent_directory)


class ClassifyViewFeature(enum.Enum):
    WITH_NOTIFICATIONS = enum.auto()


class ClassifyViewFactory:
    def __init__(
        self,
        output_json: pathlib.Path,
        notification_factory: notifications.NotificationServiceFactory,
    ):
        self.__output_json = output_json
        self.__notification_factory = notification_factory

    def build(
        self, features: FrozenSet[ClassifyViewFeature]
    ) -> cli_presenter.ClassificationView:
        _check_output(self.__output_json)
        view = json_classification_export_view.JsonClassificationExportView(
            self.__output_json
        )
        if ClassifyViewFeature.WITH_NOTIFICATIONS in features:
            notification_service = self.__notification_factory.build()
            return json_classification_export_view.NotifyClassificationView(
                view, notification_service
            )
        else:
            return view


class ReadsViewFactory:
    def __init__(
        self,
        bam_file: pathlib.Path,
        output_dir: Optional[pathlib.Path],
        notification_factory: notifications.NotificationServiceFactory,
    ):
        self.__bam_file = bam_file
        self.__output_dir = output_dir
        self.__notification_factory = notification_factory

    def build(
        self, features: FrozenSet[ClassifyViewFeature]
    ) -> cli_presenter.ReadsView:
        if self.__output_dir is None:
            raise ValueError()
        elif not self.__output_dir.exists():
            _raise_missing_directory(self.__output_dir)
        view = pysam_segment_export_view.PysamReadsView(
            self.__bam_file, self.__output_dir
        )
        if ClassifyViewFeature.WITH_NOTIFICATIONS in features:
            notification_service = self.__notification_factory.build()
            result = pysam_segment_export_view.NotifyPysamReadsView(
                view, notification_service
            )
            return result
        else:
            return view



"""Store data in domain specific objects"""

from typing import Container, Dict, Iterable, Iterator, Set, Tuple, cast

from ilmn.pelops import entities


class UnknowReadPairError(KeyError):
    def __init__(self, read_name: str) -> None:
        message = f"Unknown read pair named {read_name}"
        super().__init__(message)


class UnnamedReadPairError(ValueError):
    def __init__(self) -> None:
        message = f"Unable to store unnamed read pair"
        super().__init__(message)


class PairedReadStore:
    """Initialise, Store and Mark ReadPairs"""

    def __init__(self) -> None:
        self._reads: Dict[str, entities.ReadPair] = {}
        self._paired: Set[str] = set()
        self._spanning: Set[str] = set()
        self._split: Set[str] = set()

    def __iter__(self) -> Iterator[entities.ReadPair]:
        return iter(self._reads.values())

    def mark_as_split(self, read: entities.ReadPair) -> None:
        read_name = cast(str, read.get_name())
        self._split.add(read_name)

    def mark_as_paired(self, read: entities.ReadPair) -> None:
        read_name = cast(str, read.get_name())
        self._paired.add(read_name)

    def mark_as_spanning(self, read: entities.ReadPair) -> None:
        read_name = cast(str, read.get_name())
        self._spanning.add(read_name)

    def get_segment_count(self) -> entities.ClassifiedSegmentCount:
        paired = len(self._paired)
        split = len(self._split)
        spanning = len(self._spanning)
        result = entities.ClassifiedSegmentCount(paired, split, spanning)
        return result

    def get_read_names(self) -> Container[str]:
        return self._reads.keys()

    def get_read_pair(self, name: str) -> entities.ReadPair:
        try:
            result = self._reads[name]
        except KeyError:
            raise UnknowReadPairError(name)
        return result

    def update(self, read_pair: entities.ReadPair) -> None:
        read_name = read_pair.get_name()
        if read_name is None:
            raise UnnamedReadPairError()
        self._reads[read_name] = read_pair

    def get_spanning_reads(self) -> Iterable[entities.ReadPair]:
        return (self._reads[name] for name in self._spanning)

    def clear(self) -> None:
        self._reads.clear()
        self._split.clear()
        self._spanning.clear()
        self._paired.clear()


class SingleChromosomeRegionStore:
    """Store GenomicRegions for a particular chromosome and count repeated
    occurrences"""

    def __init__(self, chromosome_name: str):
        self._chrom_name = chromosome_name
        self._store: Dict[entities.GenomicRegion, int] = {}

    def add(self, region: entities.GenomicRegion) -> None:
        if region.chrom != self._chrom_name:
            raise KeyError(f"Unable to add region from chromosome {region.chrom}")
        elif region not in self._store.keys():
            self._store[region] = 0
        self._store[region] += 1

    def get(self) -> Iterable[Tuple[entities.GenomicRegion, int]]:
        regions = sorted(self._store.keys(), key=lambda region: region.start)
        for region in regions:
            yield region, self._store[region]

    def delete(self, region: entities.GenomicRegion) -> None:
        del self._store[region]


class RegionStore:
    """Store GenomicRegions sorted by chromosome and start position and count
    repeated occurrences"""

    def __init__(self) -> None:
        self._store: Dict[str, "SingleChromosomeRegionStore"] = {}

    def add(self, region: entities.GenomicRegion) -> None:
        if region.chrom not in self._store.keys():
            self._store[region.chrom] = SingleChromosomeRegionStore(region.chrom)
        ministore = self._store[region.chrom]
        ministore.add(region)

    def get_counted_regions(self) -> Iterable[Tuple[entities.GenomicRegion, int]]:
        """Get regions and number of observed occurrences sorted by chromosome
        and start"""

        for chrom in sorted(self._store.keys()):
            ministore = self._store[chrom]
            for region, count in ministore.get():
                yield region, count

    def get_regions(self) -> Iterator[entities.GenomicRegion]:
        """Get regions of observed occurrences sorted by chromosome and start"""
        for region, count in self.get_counted_regions():
            yield region

    def delete(self, region: entities.GenomicRegion) -> None:
        ministore = self._store[region.chrom]
        ministore.delete(region)
